% Domain

dir(0,0). dir(1,0). dir(-1,0). dir(0,1). dir(0,-1).

neighbor(X,Y,XX,YY,DX,DY) :- cell(X,Y), dir(DX,DY), XX = X+DX, YY = Y+DY,
                             cell(XX,YY), not wall(X,Y,XX,YY).

loc(x,X) :- cell(X,Y).
loc(y,Y) :- cell(X,Y).

prop(X,DX) :- loc(x,X), loc(x,XX), dir(DX,0), XX*DX < X*DX.
prop(Y,DY) :- loc(y,Y), loc(y,YY), dir(0,DY), YY*DY < Y*DY.

time(T)   :- maxsteps(T), 0 < T.
time(T-1) :- time(T), 1 < T.

at(X,Y,0) :- init(X,Y).

mino(X,Y,0) :- mino(X,Y).

% Generate

1 { at(X,Y,T) : cell(X,Y) } 1 :- time(T).

% Test

:- goal(X,Y), not time(1), not init(X,Y).
:- goal(X,Y), time(T), maxsteps(T), not at(X,Y,T).
:- goal(X,Y), time(T), at(XX,YY,T), mino(XX,YY,T), not at(X,Y,T-1).

next(X,Y,T)   :- time(T), goal(X,Y), at(X,Y,T-1).
next(XX,YY,T) :- time(T), neighbor(X,Y,XX,YY,DX,DY), at(X,Y,T-1), not goal(X,Y).
:- time(T), at(X,Y,T), not next(X,Y,T).

pos(x,X,T) :- time(T), at(X,Y,T).
pos(y,Y,T) :- time(T), at(X,Y,T).

mark(L,Z-DZ,DZ,T) :- prop(Z,DZ), pos(L,Z,T).
mark(L,Z-DZ,DZ,T) :- prop(Z,DZ), mark(L,Z,DZ,T).

pref(X,Y,T) :- mark(x,X,DX,T), neighbor(X,Y,XX,YY,DX,0).
move(X,Y,T) :- pref(X,Y,T).
move(X,Y,T) :- mark(y,Y,DY,T), neighbor(X,Y,XX,YY,0,DY).

once(XX,YY,T) :- mino(X,Y,T-1), mark(x,X,DX,T), neighbor(X,Y,XX,YY,DX,0).
once(XX,YY,T) :- mino(X,Y,T-1), mark(y,Y,DY,T), neighbor(X,Y,XX,YY,0,DY), not pref(X,Y,T).
once(X,Y,T)   :- mino(X,Y,T-1), time(T), not move(X,Y,T).

mino(XX,YY,T) :- once(X,Y,T), mark(x,X,DX,T), neighbor(X,Y,XX,YY,DX,0).
mino(XX,YY,T) :- once(X,Y,T), mark(y,Y,DY,T), neighbor(X,Y,XX,YY,0,DY), not pref(X,Y,T).
mino(X,Y,T)   :- once(X,Y,T), time(T), not move(X,Y,T).

% Display

#show at/3.
% #show mino/3.
